{% extends "!layout.html" %}

{% block rootrellink %}
        <li><a href="{{ pathto('index') }}">home</a>|&nbsp;</li>
        <li><a href="{{ pathto('news') }}">news</a>|&nbsp;</li>
        <li><a href="{{ pathto('index.html#documentation', 1) }}">documentation</a>|&nbsp;</li>
        <li><a href="{{ pathto('examples') }}">examples</a>|&nbsp;</li>
        <li><a href="{{ pathto('development') }}">development</a>|&nbsp;</li>
        <li><a href="{{ pathto('terms_overview') }}">term overview</a>|&nbsp;</li>
        <li><a href="{{ pathto('downloads') }}">downloads</a></li>
{% endblock %}

{% block relbar1 %}
<div style="background-color: white; text-align: left; padding: 10px 10px 15px 15px">
  <a href="{{ pathto('index') }}"><img src="{{
    pathto("_static/sfepy_logo_title_small.png", 1) }}" border="0" height="70px" alt="SfePy"/></a>
  <a href="http://www.zcu.cz/ntc/en/"><img src="{{
    pathto("_static/ntc_logo.png", 1) }}" align="right" border="0" height="80px" alt="NTC"/></a>
</div>
{{ super() }}
{% endblock %}

{% block sidebar1 %}{{ sidebar() }}{% endblock %}
{% block sidebar2 %}{% endblock %}

{%- block extrahead %}
{{ super() }}
<script type="text/javascript">
  var _gaq = _gaq || [];
  _gaq.push(['_setAccount', 'UA-3540566-4']);
  _gaq.push(['_setDomainName', '.sfepy.org']);
  _gaq.push(['_trackPageview']);
</script>
{% endblock %}

{% block footer %}
{{ super() }}
<div class="footer">This page uses <a href="http://analytics.google.com/">
Google Analytics</a> to collect statistics. You can disable it by blocking
the JavaScript coming from www.google-analytics.com.
<script type="text/javascript">
  (function() {
    var ga = document.createElement('script');
    ga.src = ('https:' == document.location.protocol ?
              'https://ssl' : 'http://www') + '.google-analytics.com/ga.js';
    ga.setAttribute('async', 'true');
    document.documentElement.firstChild.appendChild(ga);
  })();
</script>
</div>
{% endblock %}
