.. This (-*- rst -*-) format file contains commonly used link targets and name
   substitutions. It may be included in many files, therefore it should only
   contain link targets and name substitutions. Try grepping for "^\.\. _" to
   find plausible candidates for this list.

.. NOTE: reST targets are __not_case_sensitive__, so only one target definition
   is needed for sfepy, SfePy, Sfepy, etc...

.. sfepy links
.. _sfepy: http://sfepy.org
.. _sfepy web site: http://sfepy.org
.. _sfepy mailing list: http://groups.google.com/group/sfepy-devel
.. _our mailing list: http://groups.google.com/group/sfepy-devel
.. _sfepy issues: https://github.com/sfepy/sfepy/issues
.. _issues: https://github.com/sfepy/sfepy/issues
.. _sfepy download: http://sfepy.org/doc-devel/downloads.html
.. _download: http://sfepy.org/doc-devel/downloads.html
.. _sfepy git repository: https://github.com/sfepy/sfepy
.. _sfepy tarball: http://github.com/sfepy/sfepy/tarball/master
.. _our wiki: https://github.com/sfepy/sfepy/wiki

.. other links

.. _python: http://python.org
.. _pep-0008: http://www.python.org/dev/peps/pep-0008

.. _SciPy: http://www.scipy.org
.. _NumPy: http://www.numpy.org
.. _scikit-umfpack: http://rc.github.io/scikit-umfpack
.. _sparse matrix: http://docs.scipy.org/doc/scipy/reference/sparse.html
.. _numpy documentation guide: https://github.com/numpy/numpy/blob/master/doc/HOWTO_DOCUMENT.rst.txt
.. _docstring example: https://github.com/numpy/numpy/blob/master/doc/example.py
.. _numpydoc: http://pypi.python.org/pypi/numpydoc

.. _anaconda: https://store.continuum.io/cshop/anaconda/
.. _continuum analytics: http://www.continuum.io/
.. _enthought canopy: https://www.enthought.com/products/canopy/
.. _epd: http://www.enthought.com/products/epd.php
.. _python(x,y): http://code.google.com/p/pythonxy
.. _WinPython: http://http://winpython.sourceforge.net/

.. _cython: http://cython.org
.. _pytables: http://www.pytables.org
.. _sympy: http://sympy.org
.. _ipython: http://ipython.org
.. _pyparsing: http://sourceforge.net/projects/pyparsing
.. _umfpackpy: http://code.google.com/p/umfpackpy
.. _gmsh: http://geuz.org/gmsh
.. _medit: http://www.ann.jussieu.fr/~frey/software.html
.. _python mesh viewer: http://labs.freehackers.org/projects/pythonmeshviewer
.. _mayavi: http://code.enthought.com/projects/mayavi
.. _pysparse: http://pysparse.sourceforge.net
.. _matplotlib: http://matplotlib.sourceforge.net
.. _igakit: https://bitbucket.org/dalcinl/igakit
.. _petsc4py: https://bitbucket.org/petsc/petsc4py
.. _mpi4py: https://bitbucket.org/mpi4py/mpi4py
.. _pymetis: http://mathema.tician.de/software/pymetis

.. _git: http://git-scm.com
.. _github: http://github.com
.. _create a github account: https://github.com/signup/free
.. _gist: https://gist.github.com
.. _gitwash: http://github.com/matthew-brett/gitwash
.. _numpy gitwash: http://docs.scipy.org/doc/numpy/dev/gitwash
.. _tbaggery.com: http://tbaggery.com/2008/04/19/a-note-about-git-commit-messages.html
.. _msysgit: http://code.google.com/p/msysgit

.. _sphinx: http://sphinx.pocoo.org
.. _restructuredtext primer: http://sphinx.pocoo.org/rest.html
.. _sphinx markup constructs: http://sphinx.pocoo.org/markup/index.html
.. _docutils restructuredtext: http://docutils.sourceforge.net/rst.html

.. _tex live: http://www.tug.org/texlive

.. _new bsd license: http://www.opensource.org/licenses/bsd-license.php

.. _finite element method: http://en.wikipedia.org/wiki/Finite_element_method
.. _isogeometric analysis: http://en.wikipedia.org/wiki/Isogeometric_analysis
.. _salome download: http://www.salome-platform.org/downloads/current-version
.. _salome web site: http://www.salome-platform.org
.. _salome tutorial site: http://www.salome-platform.org/user-section/salome-tutorials
