input_name = '../examples/diffusion/time_advection_diffusion.py'
output_name_trunk = 'test_time_advection_diffusion'

from tests_basic import TestInputEvolutionary

class Test(TestInputEvolutionary):
    pass
